/*
 *  Copyright 2005-2006 The Apache Software Foundation or its licensors, as applicable.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * @author Intel, Pavel A. Ozhdikhin
 * @version $Revision: 1.18.16.4 $
 *
 */

#include "deadcodeeliminator.h"
#include "FlowGraph.h"
#include "irmanager.h"
#include "PropertyTable.h"
#include "Inst.h"
#include "Stl.h"
#include "BitSet.h"
#include "MemoryManager.h"
#include "Log.h"
#include "constantfolder.h"
#include "optimizer.h"
#include "Timer.h"

namespace Jitrino {

// Timer.h includes windows.h, which has a broken #def of min:
#undef min

DEFINE_OPTPASS_IMPL(DeadCodeEliminationPass, dce, "Dead Code Elimination")

void
DeadCodeEliminationPass::_run(IRManager& irm) {
    DeadCodeEliminator dce(irm);
    {
        static Timer* justDceTimer = NULL;
        PhaseTimer t(justDceTimer, "opt::dce::justDce"); 
        dce.eliminateDeadCode(false);
    }
}

DEFINE_OPTPASS_IMPL(UnreachableCodeEliminationPass, uce, "Unreachable Code Elimination")

void
UnreachableCodeEliminationPass::_run(IRManager& irm) {
    DeadCodeEliminator dce(irm);
    {
        static Timer* justUceTimer = NULL;
        PhaseTimer t(justUceTimer, "opt::dce::justUce"); 
        dce.eliminateUnreachableCode();
    }
    bool fixup_ssa = irm.getParameterTable().lookupBool("opt::fixup_ssa", true);
    if (irm.getInSsa() && fixup_ssa) {
        static Timer* edcFixupTimer = NULL;
        PhaseTimer t(edcFixupTimer, "opt::dce::edcFixup"); 
        fixupSsa(irm);
    }
}

DEFINE_OPTPASS_IMPL(PurgeEmptyNodesPass, purge, "Empty Node Removal")

void
PurgeEmptyNodesPass::_run(IRManager& irm) {
    irm.getFlowGraph().purgeEmptyNodes(false);
}

//
// Returns true if an instruction is a non-essential instruction
// Essential instructions are exception throwing instructions, calls,
// returns, stores, and branches.  Once we have control dependence 
// information branches are non-essential.
//
static bool
isNonEssential(Inst* inst) {
    Operation operation = inst->getOperation();
    if (operation.isNonEssential()) {
        return true;
    } else {
        if (inst->getOpcode() == Op_StVar) {
            // StVar is non-essential if the variable to which it stores is in SSA form
            return inst->getDst()->isSsaVarOpnd();
        }
        return false;
    }
}

static uint8
getBitWidth(Type::Tag tag)
{
    switch (tag) {
    case Type::Void: return 0;
    case Type::Boolean: return 1;
    case Type::Char: return 16;
    case Type::Int8: return 8;
    case Type::Int16: return 16;
    case Type::Int32: return 32;
    case Type::Int64: return 64;
    case Type::UInt8: return 8;
    case Type::UInt16: return 16;
    case Type::UInt32: return 32;
    case Type::UInt64: return 64;

    default: return 0xff; // uses all;
    }        
}

// returns the number of bits of the given opnd_num which is used,
// given that dst_bits of the dst are used
static uint8
usesBitsOfOpnd(uint8 dst_bits, Inst* inst, uint32 opnd_num) {
    uint8 opwidth = getBitWidth(inst->getType());
    Opnd *opnd = inst->getSrc(opnd_num);
    uint8 opndwidth = getBitWidth(opnd->getType()->tag);
    uint8 res = opndwidth;
    switch (inst->getOpcode()) {
    case Op_Add:   case Op_Mul:    case Op_Sub:
        res = ::std::min(dst_bits, opwidth); break;
    case Op_TauDiv:   case Op_TauRem:    
        res = dst_bits; break;
    case Op_Neg:   
        res = ::std::min(dst_bits, opwidth); break;
    case Op_MulHi:
        res = opwidth; break;
    case Op_Min:   case Op_Max:   case Op_Abs:
    case Op_And:   case Op_Or:    case Op_Xor:    case Op_Not:
        res = ::std::min(dst_bits, opwidth); break;
    case Op_Select:
        if (opnd_num == 0) res = opndwidth;
        else res = dst_bits;
        break;
    case Op_Conv:
        res = ::std::min(dst_bits, opwidth); break;
    case Op_Shladd:  
        {
            switch (opnd_num) {
            case 0:
                {
                    Opnd *shiftAmount = inst->getSrc(1);
                    int32 shiftby;
                    bool isconst = 
                        ConstantFolder::isConstant(shiftAmount->getInst(), 
                                                   shiftby);
                    if( !(isconst && (shiftby > 0)) ) assert(0);
                    res = ::std::min(dst_bits, opwidth)-(uint8)shiftby;
                    break;
                }
            case 1:
                {
                    res = opndwidth;
                    break;
                }
            case 2:
                {
                    res = ::std::min(dst_bits, opwidth);
                    break;
                }
            default:
                assert(0);
            }
            break;
        }        
    case Op_Shl:
        {
            switch (opnd_num) {
            case 0:
                {
                    Opnd *shiftAmount = inst->getSrc(1);
                    int32 shiftby;
                    bool isconst 
                        = ConstantFolder::isConstant(shiftAmount->getInst(), 
                                                     shiftby);
                    if (isconst && shiftby > 0) {
                        res = ::std::min(dst_bits, opwidth)-(uint8)shiftby;
                        break;
                    }
                }
            case 1:
                {
                    res = opndwidth;
                }
                break;
            default:
                assert(0);
            };
            break;
        }
    case Op_Shr:   
        res = opndwidth; break;
    case Op_Cmp:    case Op_Cmp3:
    case Op_Copy:
        res = ::std::min(dst_bits, opwidth); break;
    case Op_LdVar:
        res = dst_bits; break;
    case Op_Phi:
        res = dst_bits; break;
    case Op_StVar: 
        if (inst->getDst()->isSsaVarOpnd()) {
            res = dst_bits;
        } else {
            res = opndwidth; // can't track value accurately, be conservative
        }
        break;
    case Op_TauStElem:
    case Op_TauStRef: 
        {
            if (opnd_num == 0) { // value to be stored, trim to opwidth
                res = opwidth;
            }
        }
        break;
    case Op_TauStInd:
    case Op_TauStField: 
        {
            if (opnd_num == 0) { // value to be stored, trim to opwidth
                res = opwidth;
            }
        }
        break;
    case Op_TauStStatic:
        {
            if (opnd_num == 0) { // value to be stored, trim to opwidth
                res = opwidth;
            }
        }
        break;
    default:
        {
            res = opndwidth;
        }
    }
    uint8 res1 = ::std::min(res, opndwidth);
    return res1;
}

DeadCodeEliminator::DeadCodeEliminator(IRManager& irm)
    : irManager(irm), flowGraph(irm.getFlowGraph()), returnOpnd(irm.getReturnOpnd()) {
    preserveCriticalEdges = irManager.getParameterTable().lookupBool("opt::dce::preserve_critical_edges", true);
}

Opnd*
DeadCodeEliminator::findDefiningTemp(Opnd* var) {
    SsaVarOpnd* ssaVarOpnd = var->asSsaVarOpnd();
    if(ssaVarOpnd == NULL) {
        Log::cat_opt_dc()->debug << "Nothing found: Not SSA" << ::std::endl;
        return NULL; // not an SsaVarOpnd
    }
    Inst* inst = ssaVarOpnd->getInst();
    assert(inst->getNode());
    if(inst->getOpcode() == Op_StVar) {
        // propagate src of copy recursively
        copyPropagate(inst);
        if(Log::cat_opt_dc()->isDebugEnabled()) {
            Log::out() << "Found: ";
            Inst* def = inst->getSrc(0)->getInst();
            if(def != NULL) {
                def->print(Log::out());
            }
            else { 
                Log::out() << "Dangling operand "; inst->getSrc(0)->print(Log::out()); 
            }
            Log::out() << ::std::endl;
        }
        return inst->getSrc(0);
    } else if(inst->getOpcode() == Op_Phi) {

        Opnd* tmp = NULL;
        uint32 n = inst->getNumSrcOperands();
        for(uint32 j=0; j < n; ++j) {
            Opnd* src = inst->getSrc(j);
            Inst* srcInst = src->getInst();
            assert(srcInst->getNode());
            if(srcInst == NULL || srcInst->getOpcode() != Op_StVar) {
                tmp = NULL;
                break;
            }
            assert(srcInst->getOpcode() == Op_StVar);
            if(tmp == NULL) {
                assert(j == 0);
                tmp = srcInst->getSrc(0);
            } else if(tmp != srcInst->getSrc(0)) {
                tmp = NULL;
                break;
            }
        }

        return tmp;
    }
    Log::cat_opt_dc()->debug << "Nothing found: Not stVar or phi instruction" << ::std::endl;
    return NULL;
}


Opnd*
DeadCodeEliminator::copyPropagate(Opnd* opnd) {
    SsaOpnd* src = opnd->asSsaOpnd();
    if (src == NULL)
        return opnd;   // src was not a SsaOpnd so it cannot be propagated
    Inst* srcInst = src->getInst();
    assert(srcInst->getNode());
    if (srcInst == NULL)
        return opnd;
    if (srcInst->getOpcode() == Op_Copy) {
        // propagate src of copy recursively
        copyPropagate(srcInst);
        return srcInst->getSrc(0);
    } else if (srcInst->getOpcode() == Op_LdVar) {
        // propagate src of LdVar if src SsaVarOpnd comes from a StVar
        Opnd* tmp = findDefiningTemp(srcInst->getSrc(0));
        if(tmp != NULL) {
            return tmp;
        } else {
            return opnd;
        }
    } else if (srcInst->getOpcode() == Op_Phi) {
        // propagate src of degenerate phi (phi with only one arg).
        if(srcInst->getNumSrcOperands() != 1)
            return opnd;
        // propagate src of copy recursively
        copyPropagate(srcInst);
        // reset inst's source to propagates src
        return srcInst->getSrc(0);
    }
    return opnd;
}

void
DeadCodeEliminator::copyPropagate(Inst* inst) {
    uint32 numSrcs = inst->getNumSrcOperands();
    for (uint32 i=0; i<numSrcs; i++) {
        inst->setSrc(i, copyPropagate(inst->getSrc(i)));
    }
}

//
// marks an instruction as live and add its sources to the work set
//
static void
markLiveInst1(Inst* inst,
              InstDeque& workSet,
              BitSet& usefulInstSet,
              BitSet& usefulVarSet,
              uint32 minInstId,
              uint32 maxInstId) {
    //
    // add instruction's sources to the work list
    //
    assert(inst);
    assert(inst->getNode());

    uint8 dstWidth = 255;

    if (Log::cat_opt_dc()->isDebugEnabled()) {
        Log::out() << "Found dstwidth of " << (int) dstWidth
                   << " for inst ";
        inst->print(Log::out());
        Log::out() << ::std::endl;
    }

    for (uint32 i=0; i<inst->getNumSrcOperands(); i++) {
        Opnd* src = inst->getSrc(i);
        // only follow ssa use-def links
        SsaOpnd* srcSsaOpnd = src->asSsaOpnd();
        if (srcSsaOpnd != NULL) {
            // use of an ssa operand
            Inst* def = srcSsaOpnd->getInst();
            if(def == NULL)
                continue;
            assert(def != NULL);

            assert(def->getNode());
            uint8 opndWidth = 255;
            uint32 defId = def->getId();

            if (usefulInstSet.setBit(defId-minInstId, true)) {
                // was already marked as live, check old width.
                uint8 oldOpndWidth = 255;

                if (Log::cat_opt_dc()->isDebugEnabled()) {
                    Log::out() << "Found dstwidth of " << (int) oldOpndWidth
                               << " for inst ";
                    def->print(Log::out());
                    Log::out() << ::std::endl;
                }

                if (opndWidth > oldOpndWidth) {
                    if (Log::cat_opt_dc()->isDebugEnabled()) {
                        Log::out() << "Setting dstwidth to " << (int) opndWidth
                                   << " for inst ";
                        def->print(Log::out());
                        Log::out() << ::std::endl;
                    }
                    workSet.pushFront(def);                    
                }
            } else {
                if (Log::cat_opt_dc()->isDebugEnabled()) {
                    Log::out() << "Setting dstwidth to " << (int) opndWidth
                               << " for inst ";
                    def->print(Log::out());
                    Log::out() << ::std::endl;
                }
                workSet.pushFront(def);
            }
            // mark ssavaropnds as not dead
            SsaVarOpnd* ssaVarOpnd = src->asSsaVarOpnd();
            if (ssaVarOpnd != NULL) {
                // mark var as useful
                usefulVarSet.setBit(ssaVarOpnd->getVar()->getId(),true);
            }
        } else {
            VarOpnd* srcVarOpnd = src->asVarOpnd();
            if (srcVarOpnd != NULL) {
                // use of a var opnd (ldvar or ldvara)
                // mark var as useful
                usefulVarSet.setBit(srcVarOpnd->getId(),true);
            }
        }
    }
}

//
// Deletes instructions that are not marked as useful
//
void
DeadCodeEliminator::sweepInst1(CFGNode* node, 
                               Inst* inst,
                               BitSet& usefulInstSet,
                               BitSet& usefulVarSet,
                               uint32 minInstId,
                               uint32 maxInstId,
                               bool canRemoveStvars) {
    assert(inst);
    assert(inst->getNode());
    uint32 instId = inst->getId();
    if (inst->isMethodMarker()) {
        // don't remove a method marker, but remove opcode if it is dead
        if (inst->getNumSrcOperands() > 0) {
            Opnd *opnd = inst->getSrc(0);
            if (opnd && !opnd->isNull()) {
                Inst *srcInst = opnd->getInst();
                assert(srcInst);
                assert(srcInst->getNode());
                uint32 srcId = srcInst->getId();
                // instructions are only added during analysis for
                // live instructions, so if it's not in the map, 
                // assume it is live
                if ((minInstId <= srcId) &&
                    (srcId < maxInstId) &&
                    (usefulInstSet.getBit(srcId-minInstId) == false)) {
                    MethodMarkerInst *mmi = inst->asMethodMarkerInst();
                    mmi->removeOpnd();
                }
            }
        }            
    } else if ((minInstId <= instId) && (instId < maxInstId) &&
               usefulInstSet.getBit(instId-minInstId) == false) {
        // inst is a useless instruction
        if (Log::cat_opt_dc()->isDebugEnabled()) {
            Log::out() << "Useless inst found, removing: ";
            inst->print(Log::out());
            Log::out() << ::std::endl;
            Log::out() << " instId=" << (int)instId;
            Log::out() << ", minInstId=" << (int)minInstId;
            Log::out() << ", maxInstId=" << (int)maxInstId;
            Log::out() << ::std::endl;
        }
        inst->unlink();
        if (inst->isBranch()) {
            assert(0);
        } else if (inst->isSwitch()) {
            assert(0);
        } else if (inst->getOperation().canThrow()) {
            flowGraph.removeEdge(node->getExceptionEdge());
        }
        return;
    }
    //
    // delete stores to useless variables
    //
    if (canRemoveStvars && inst->getOpcode() == Op_StVar) {
        VarOpnd* dstVarOpnd = inst->getDst()->asVarOpnd();
        if (dstVarOpnd == NULL)
            return;
        if (usefulVarSet.getBit(dstVarOpnd->getId()) == false) {
            //
            // store to a useless variable
            //
            if (Log::cat_opt_dc()->isDebugEnabled()) {
                Log::out() << "Removing store to useless var: ";
                inst->print(Log::out());
                Log::out() << ::std::endl;
            }
            inst->unlink();
            return;
        }       
    }
}

//
// marks an instruction as live and add its sources to the work set
//
static void
markLiveInst(Inst* inst,
             InstDeque& workSet,
             BitSet& usefulInstSet,
             BitSet& usefulVarSet,
             uint8 *usedInstWidth,
             uint32 minInstId,
             uint32 maxInstId) {
    //
    // add instruction's sources to the work list
    //
    assert(inst);
    assert(inst->getNode());
    uint32 instId = inst->getId();
    assert(usedInstWidth);
    assert((instId >= minInstId) && (instId < maxInstId));
    uint8 dstWidth = usedInstWidth[instId-minInstId];

    if (Log::cat_opt_dc()->isDebugEnabled()) {
        Log::out() << "Found dstwidth of " << (int) dstWidth
                   << " for inst ";
        inst->print(Log::out());
        Log::out() << ::std::endl;
    }

    for (uint32 i=0; i<inst->getNumSrcOperands(); i++) {
        Opnd* src = inst->getSrc(i);
        // only follow ssa use-def links
        SsaOpnd* srcSsaOpnd = src->asSsaOpnd();
        if (srcSsaOpnd != NULL) {
            // use of an ssa operand
            Inst* def = srcSsaOpnd->getInst();
            if(def == NULL)
                continue;
            assert(def != NULL);

            uint8 opndWidth = usesBitsOfOpnd(dstWidth, inst, i);
            assert(def);
            assert(def->getNode());
            uint32 defId = def->getId();

            if (!((minInstId <= defId) && (defId < maxInstId))) {
                // this instruction is out of the region, skip it.
                
                if (Log::cat_opt_dc()->isDebugEnabled()) {
                    Log::out() << "Skipping outside-region inst ";
                    def->print(Log::out());
                    Log::out() << ::std::endl;
                }
            } else if (usefulInstSet.setBit(defId-minInstId, true)) {
                // was already marked as live, check old width.
                uint8 oldOpndWidth = usedInstWidth[defId-minInstId];

                if (Log::cat_opt_dc()->isDebugEnabled()) {
                    Log::out() << "Found dstwidth of " << (int) oldOpndWidth
                               << " for inst ";
                    def->print(Log::out());
                    Log::out() << ::std::endl;
                }

                if (opndWidth > oldOpndWidth) {
                    if (Log::cat_opt_dc()->isDebugEnabled()) {
                        Log::out() << "Setting dstwidth to " << (int) opndWidth
                                   << " for inst ";
                        def->print(Log::out());
                        Log::out() << ::std::endl;
                    }
                    usedInstWidth[defId-minInstId] = opndWidth;
                    workSet.pushFront(def);                    
                }
            } else {
                if (Log::cat_opt_dc()->isDebugEnabled()) {
                    Log::out() << "Setting dstwidth to " << (int) opndWidth
                               << " for inst ";
                    def->print(Log::out());
                    Log::out() << ::std::endl;
                }
                usedInstWidth[defId-minInstId] = opndWidth;
                workSet.pushFront(def);
            }
            // mark ssavaropnds as not dead
            SsaVarOpnd* ssaVarOpnd = src->asSsaVarOpnd();
            if (ssaVarOpnd != NULL) {
                // mark var as useful
                usefulVarSet.setBit(ssaVarOpnd->getVar()->getId(),true);
            }
        } else {
            VarOpnd* srcVarOpnd = src->asVarOpnd();
            if (srcVarOpnd != NULL) {
                // use of a var opnd (ldvar or ldvara)
                // mark var as useful
                usefulVarSet.setBit(srcVarOpnd->getId(),true);
            }
        }
    }
}

//
// Deletes instructions that are not marked as useful
//
void
DeadCodeEliminator::sweepInst(CFGNode* node, 
                              Inst* inst,
                              BitSet& usefulInstSet,
                              BitSet& usefulVarSet,
                              uint8 *usedInstWidth,
                              uint32 minInstId,
                              uint32 maxInstId,
                              bool canRemoveStvars) {
    assert(usedInstWidth);
    assert(inst);
    assert(inst->getNode());
    uint32 instId = inst->getId();
    if (inst->isMethodMarker()) {
        // don't remove a method marker, but remove opcode if it is dead
        if (inst->getNumSrcOperands() > 0) {
            Opnd *opnd = inst->getSrc(0);
            if (opnd && !opnd->isNull()) {
                Inst *srcInst = opnd->getInst();
                assert(srcInst);
                assert(srcInst->getNode());
                uint32 srcId = srcInst->getId();
                // instructions are only added during analysis for
                // live instructions, so if it's not in the map, 
                // assume it is live
                if ((minInstId <= srcId) && (srcId < maxInstId) &&
                    (usefulInstSet.getBit(srcId-minInstId) == false)) {
                    MethodMarkerInst *mmi = inst->asMethodMarkerInst();
                    mmi->removeOpnd();
                }
            }
        }            
    } else if ((minInstId <= instId) && (instId < maxInstId) &&
               usefulInstSet.getBit(instId-minInstId) == false) {
        // inst is a useless instruction
        if (Log::cat_opt_dc()->isDebugEnabled()) {
            Log::out() << "Useless inst found, removing: ";
            inst->print(Log::out());
            Log::out() << ::std::endl;
            Log::out() << " instId=" << (int)instId;
            Log::out() << ", minInstId=" << (int)minInstId;
            Log::out() << ", maxInstId=" << (int)maxInstId;
            Log::out() << ::std::endl;
        }
        inst->unlink();
        if (inst->isBranch()) {
            assert(0);
        } else if (inst->isSwitch()) {
            assert(0);
        } else if (inst->getOperation().canThrow()) {
            flowGraph.removeEdge(node->getExceptionEdge());
        }
        return;
    } else {
        switch (inst->getOpcode()) {
        case Op_Conv:
            {
                Type::Tag instType = inst->getType();
                
                if ((!Type::isInteger(instType)) ||
                    ((inst->getOverflowModifier() != Overflow_None) &&
                     (inst->getExceptionModifier() != Exception_Never))) 
                    break;

                Type *srcType = inst->getSrc(0)->getType();
                if (!srcType->isInteger())
                    break;
        
                uint8 convSize = getBitWidth(instType);
                uint8 srcSize = getBitWidth(srcType->tag);
                uint8 dstSize = getBitWidth(inst->getDst()->getType()->tag);

                if (dstSize != srcSize) {
                    // it's not just a sign-extension, we can't remove it
                    break;
                }

                assert((minInstId <= instId) && (instId < maxInstId));
                uint8 usedDstSize = usedInstWidth[instId-minInstId];

                if (Log::cat_opt_dc()->isDebugEnabled()) {
                    Log::out() << "Found dstwidth of " << (int) usedDstSize
                               << " for inst ";
                    inst->print(Log::out());
                    Log::out() << ::std::endl;
                }
                
                if ((srcSize >= convSize) &&
                    (usedDstSize <= convSize)) {
                    // convert to a copy.
                    if (Log::cat_opt_dc()->isDebugEnabled()) {
                        Log::out() << "Eliminating redundant conv ";
                        inst->print(Log::out());
                        Log::out() << ::std::endl;
                    }
                    InstFactory &factory = irManager.getInstFactory();
                    Inst *copyInst =
                        factory.makeCopy(inst->getDst(), inst->getSrc(0));
                    copyInst->insertAfter(inst);
                    inst->unlink();
                    // copyInst doesn't appear in bitmaps, but we won't walk
                    // over it, so it is only tested for a MethodMarker Inst 
					// operand above, and we are careful there
                }
            }
            break;
        default:
            break;
        }
    }
    //
    // delete stores to useless variables
    //
    if (canRemoveStvars && inst->getOpcode() == Op_StVar) {
        VarOpnd* dstVarOpnd = inst->getDst()->asVarOpnd();
        if (dstVarOpnd == NULL)
            return;
        if (usefulVarSet.getBit(dstVarOpnd->getId()) == false) {
            //
            // store to a useless variable
            //
            if (Log::cat_opt_dc()->isDebugEnabled()) {
                Log::out() << "Removing store to useless var: ";
                inst->print(Log::out());
                Log::out() << ::std::endl;
            }
            inst->unlink();
            return;
        }       
    }
}

//
// Eliminate unreachable code
//
bool
DeadCodeEliminator::eliminateUnreachableCode() {
    MemoryManager memManager(1000,"DeadCodeEliminator::eliminateUnreachableCode");

    // Purge unreachable nodes from CFG.
    StlVector<CFGNode*> unreachableNodes(memManager);
    flowGraph.purgeUnreachableNodes((StlVector<ControlFlowNode*>&) unreachableNodes);   

    // If no unreachable nodes, return
    if(unreachableNodes.empty())
        return false;

    // Clear the set of unreachable definitions.
    StlHashSet<uint32> unreachableDefSet(memManager);

    StlVector<CFGNode*>::iterator niter;
    for(niter = unreachableNodes.begin(); niter != unreachableNodes.end(); ++niter) {
        CFGNode* node = *niter;

        // Find unreachable defs.  Note, only SsaVarOpnds can be used 
        // in later reachable nodes.
        Inst* headInst = node->getFirstInst();
        assert(headInst->getNode());
        for (Inst* inst = headInst->next();inst!=headInst;inst=inst->next()) {
            assert(inst->getNode());
            SsaVarOpnd* dst = inst->getDst()->asSsaVarOpnd();
            if(dst != NULL) {
                assert(inst->getOpcode() == Op_StVar || inst->getOpcode() == Op_Phi);
                unreachableDefSet.insert(dst->getId());
            }
        }
    }

    // If no unreachable defs to cleanup, just return.
    if(unreachableDefSet.empty())
        return true;


    // Cleanup up phi instructions.
    const CFGNodeDeque& nodes = flowGraph.getNodes();
    CFGNodeDeque::const_iterator niter2;
    for(niter2 = nodes.begin(); niter2 != nodes.end(); ++niter2) {
        CFGNode* node = *niter2;
        Inst* headInst = node->getFirstInst();
        assert(headInst->getNode());
        for (Inst* inst = headInst->next();inst!=headInst;inst=inst->next()) {
            assert(inst->getNode());
            if(inst->getOpcode() == Op_Phi) {
#ifndef NDEBUG
                SsaVarOpnd* dst = inst->getDst()->asSsaVarOpnd();
                assert(dst!=NULL);
#endif
				uint32 numSrc = inst->getNumSrcOperands();
                uint32 numKill = 0;
                for(uint32 i = 0; i < numSrc; ++i) {
                    SsaVarOpnd* src = inst->getSrc(i)->asSsaVarOpnd();
                    assert(src != NULL);
                    if(unreachableDefSet.find(src->getId()) != unreachableDefSet.end()) {
                        // Purge this operand.
                        ++numKill;
                    } else {
                        // Shift down over purged operands.
                        assert(numKill <= i);
                        if(numKill > 0)
                            inst->setSrc(i-numKill, src);
                    }
                }
                if(numKill > 0)
                    ((PhiInst*) inst)->setNumSrcs(numSrc-numKill);
            }
        }
    }
    return true;
}

static Timer *dcePhase1Timer = 0; // not thread-safe
static Timer *dcePhase2Timer = 0; // not thread-safe
static Timer *dcePhase3Timer = 0; // not thread-safe
static Timer *dcePhase4Timer = 0; // not thread-safe
static Timer *dcePhase5Timer = 0; // not thread-safe
static Timer *dcePhase6Timer = 0; // not thread-safe

//
// Performs dead code elimination
//
void
DeadCodeEliminator::eliminateDeadCode(bool keepEmptyNodes) {
    // user should call eliminateUnreachableCode() first

    uint32 minInstId = irManager.getMinimumInstId();
    uint32 maxInstId = irManager.getInstFactory().getNumInsts();
    uint32 numInsts = maxInstId - minInstId;
    uint32 numOpnds = irManager.getOpndManager().getNumVarOpnds();
    MemoryManager memManager(numInsts*20,
                             "DeadCodeEliminator::eliminateDeadCode");
    InstDeque workSet(memManager);
    CFGVector nodes(memManager,flowGraph.getMaxNodeId());
    flowGraph.getNodesPostOrder(nodes);
    CFGVector::ReverseIterator niter;

    // set of useful instructions & variables; initially everything is useless
    BitSet usefulInstSet(memManager,numInsts+1);
    BitSet usefulVarSet(memManager,numOpnds+1);
    OptimizerFlags& optimizerFlags = *irManager.getCompilationContext()->getOptimizerFlags();
    uint8 *usedInstWidth = optimizerFlags.dce2 ? new (memManager) uint8[numInsts] : 0;

    {
        PhaseTimer t(dcePhase1Timer, "opt::dce::phase1"); 
        
    //
    // first propagate copies and initialize the work list with 
    // essential instructions
    //
    if (usedInstWidth) {
        for (niter = nodes.rbegin(); niter != nodes.rend(); ++niter) {
            CFGNode* node = *niter;
            Inst* headInst = node->getFirstInst();
            assert(headInst->getNode());
            for (Inst* inst = headInst->next();inst!=headInst;inst=inst->next()) {
                assert(inst->getNode());
                // copy propagate all sources of this instruction
                copyPropagate(inst);
                // For inlined methods, the last instruction of the epilog copies the return value out.
                if (isNonEssential(inst) == false || (returnOpnd != NULL && returnOpnd == inst->getDst())) {
                    // add essential instruction to work list
                    assert(inst != NULL);
                    workSet.pushBack(inst);

                    assert(inst);
                    assert(inst->getNode());
                    uint32 instId = inst->getId();
                    usefulInstSet.setBit(instId-minInstId, true);
                    if (usedInstWidth) {
                        uint8 usedWidth = getBitWidth(inst->getType());
                        if (Log::cat_opt_dc()->isDebugEnabled()) {
                            Log::out() << "Setting dstwidth to " << (int) usedWidth
                                       << " for inst ";
                            inst->print(Log::out());
                            Log::out() << ::std::endl;
                        }
                        assert((minInstId <= instId) && (instId < maxInstId));
                        usedInstWidth[instId-minInstId]= usedWidth;
                    }
                }
            }
        }
    } else {
        for (niter = nodes.rbegin(); niter != nodes.rend(); ++niter) {
            CFGNode* node = *niter;
            Inst* headInst = node->getFirstInst();
            assert(headInst->getNode());
            for (Inst* inst = headInst->next();inst!=headInst;inst=inst->next()) {
                assert(inst->getNode());
                // copy propagate all sources of this instruction
                copyPropagate(inst);
                // For inlined methods, the last instruction of the epilog copies the return value out.
                if (isNonEssential(inst) == false || (returnOpnd != NULL && returnOpnd == inst->getDst())) {
                    // add essential instruction to work list
                    assert(inst != NULL);
                    workSet.pushBack(inst);

                    assert(inst);
                    assert(inst->getNode());
                    uint32 instId = inst->getId();
                    usefulInstSet.setBit(instId-minInstId, true);
                }
            }
        }
    }
    }

    {
        PhaseTimer t(dcePhase2Timer, "opt::dce::phase2"); 

    //
    // Iteratively mark all useful instructions by computing slices
    //
    if (usedInstWidth) {
        while (!workSet.isEmpty()) {
            markLiveInst(workSet.popFront(),workSet,usefulInstSet,usefulVarSet,
                         usedInstWidth, minInstId, maxInstId);
        }
    } else {
        while (!workSet.isEmpty()) {
            markLiveInst1(workSet.popFront(),workSet,
                          usefulInstSet, usefulVarSet, minInstId, maxInstId);
        }
    }
    }

    {
        const bool canRemoveStVars = (irManager.getParent() == NULL); // we can remove a useless var
        PhaseTimer t(dcePhase3Timer, "opt::dce::phase3"); 
        //
        // Now cleanup all the dead code
        //
        if (usedInstWidth) {
            for (niter = nodes.rbegin(); niter != nodes.rend(); ++niter) {
                CFGNode* node = *niter;
                Inst* headInst = node->getFirstInst();
                assert(headInst->getNode());
                for (Inst* inst = headInst->next(); inst != headInst; ) {
                    assert(inst->getNode());
                    Inst *nextInst = inst->next();
                    sweepInst(node, inst,usefulInstSet,usefulVarSet,usedInstWidth,minInstId,maxInstId,canRemoveStVars);
                    inst = nextInst;
                }
            }
        } else {
            for (niter = nodes.rbegin(); niter != nodes.rend(); ++niter) {
                CFGNode* node = *niter;
                Inst* headInst = node->getFirstInst();
                assert(headInst->getNode());
                for (Inst* inst = headInst->next(); inst != headInst; ) {
                    assert(inst->getNode());
                    Inst *nextInst = inst->next();
                    sweepInst1(node, inst,usefulInstSet,usefulVarSet,minInstId,maxInstId,canRemoveStVars);
                    inst = nextInst;
                }
            }
        }
    }

    {
        const bool canUnlink = (irManager.getParent() == NULL); // we can unlink an unused var

        PhaseTimer t(dcePhase4Timer, "opt::dce::phase4"); 
        //
        // delete dead variables
        //
        OpndManager &opndManager = irManager.getOpndManager();
        VarOpnd *varOpnd0 = opndManager.getVarOpnds();
        if (varOpnd0) {
            VarOpnd *varOpnd = varOpnd0;
            // first check whether head needs to be deleted
            while (varOpnd == varOpnd0) {
                VarOpnd *nextVar = varOpnd->getNextVarOpnd();
                if (nextVar == varOpnd) nextVar = 0; // this should make us stop
                if (usefulVarSet.getBit(varOpnd->getId()) == true){
                    // live variable
                    varOpnd->setDeadFlag(false);  // mark as live
                } else {
                    // dead variable
                    varOpnd->setDeadFlag(true); // mark as dead
                    
                    if (canUnlink) {
                        if (Log::cat_opt_dc()->isDebugEnabled()) {
                            Log::out() << "removing dead VarOpnd ";
                            varOpnd->print(Log::out());
                            Log::out() << ::std::endl;
                        }
                        opndManager.deleteVar(varOpnd); // remove from list
                    }
                    
                    // get the new head; this should usually be nextVar, but may be 0.
                    varOpnd0 = opndManager.getVarOpnds();
                    if (!nextVar && !varOpnd0) nextVar = varOpnd; // if it is 0, make sure we still stop
                }
                varOpnd = nextVar;
            };
            // we may have deleted all vars already, be careful here.
            if (varOpnd0) {
                varOpnd = varOpnd0->getNextVarOpnd();
                while (varOpnd != varOpnd0) {
                    VarOpnd *nextVar = varOpnd->getNextVarOpnd();
                    if (usefulVarSet.getBit(varOpnd->getId()) == true){
                        // live variable
                        varOpnd->setDeadFlag(false);  // mark as live
                    } else {
                        // dead variable

                        varOpnd->setDeadFlag(true); // mark as dead
                        if (canUnlink) {
                            if (Log::cat_opt_dc()->isDebugEnabled()) {
                                Log::out() << "removing dead VarOpnd ";
                                varOpnd->print(Log::out());
                                Log::out() << ::std::endl;
                            }
                            opndManager.deleteVar(varOpnd); // remove from list
                        }
                    }
                    varOpnd = nextVar;
                }
            }
        }
    }

    {
        PhaseTimer t(dcePhase6Timer, "opt::dce::phase6"); 
        
        flowGraph.mergeAdjacentNodes();       
    }
    {
        PhaseTimer t(dcePhase5Timer, "opt::dce::phase5"); 
        if (!keepEmptyNodes) {
            //
            // Purge empty nodes.
            //
            flowGraph.purgeEmptyNodes(preserveCriticalEdges);
        }
    }

}

} //namespace Jitrino 
